/*
 * Copyright (c) 2017 Objective Systems, Inc.
 *
 * See the COPYING.txt file in this subdirectory for conditions on use
 * of this software.  This notice applies to software only in this
 * subdirectory (oortsrc) and not to other software that may be a
 * a part of this distribution.
 *
 *****************************************************************************/
/**
 * @file ooTimer.h
 * Timer structures and functions.
 */
#ifndef _OOTIMER_H_
#define _OOTIMER_H_

#include "oortsrc/ooSocket.h"   /* needed for timeval */

struct _OOTimer;

typedef int (*OOTimerCbFunc)(struct _OOTimer *pTimer, void *data);

typedef struct _OOTimer {
   struct timeval expireTime, timeout;
   void*	cbData;
   OOBOOL       reRegister;

   /* Callback functions */
   OOTimerCbFunc timeoutCB;
} OOTimer;

#ifdef __cplusplus
extern "C" {
#endif
/**
 * This function computes the relative expiration time from the current
 * time for the given timer object.
 *
 * @param pTimer       Pointer to timer object.
 */
EXTERN void ooTimerComputeExpireTime (OOTimer* pTimer);

/**
 * This function creates and initializes a new timer object.
 *
 * @param cb           Timer callback function.
 * @param deltaSecs    Time in seconds to timer expiration.
 * @param data         Callback user data argument.
 * @param reRegister   Should timer be re-registered after it expires?
 * @return             Pointer to created timer object.
 */
EXTERN OOTimer* ooTimerCreate
(OOTimerCbFunc cb, OOUINT32 deltaSecs, void *data, OOBOOL reRegister);

/**
 * This function deletes the given timer object.
 *
 * @param pTimer       Pointer to timer object.
 */
EXTERN void ooTimerDelete (OOTimer* pTimer);

/**
 * This function checks a timer to determine if it is expired.
 *
 * @param pTimer       Pointer to timer object.
 * @return             True if timer expired, false if not.
 */
EXTERN OOBOOL ooTimerExpired (OOTimer* pTimer);

/**
 * This function loops through the global timer list and fires all
 * expired timers by calling the registered callback functions.
 */
EXTERN void ooTimerFireExpired (void);

/**
 * This function inserts the given timer object into the correct
 * chronological position in the global timer list.
 *
 * @param pTimer       Pointer to timer object.
 * @return             Index to position where inserted in list.
 */
EXTERN int ooTimerInsertEntry (OOTimer* pTimer);

/**
 * This function calculates the relative time from the current time
 * that the first timer in global timer list will expire.
 *
 * @param ptimeout      timeval structure to receive timeout value.
 * @return              ptimeout
 */
EXTERN struct timeval* ooTimerNextTimeout (struct timeval* ptimeout);

/**
 * This function resets the given timer object if its reregister flag
 * is set.  Otherwise, it is deleted.
 *
 * @param pTimer       Pointer to timer object.
 */
EXTERN void ooTimerReset (OOTimer* pTimer);

#ifdef __cplusplus
}
#endif
#endif
